import mkdocs
from mkdocs.config import config_options, Config
from mkdocs.structure.nav import Section
from mkdocs.structure.files import Files
from mkdocs.utils import meta as Meta
from os import path, rename
import re
import urllib.parse

class DirectoryMetaInfoPlugin(mkdocs.plugins.BasePlugin):
    # html link pattern
    LINK_PATTERN = re.compile(r'<a\s.*?href="(?P<href>.*?)">', re.MULTILINE | re.DOTALL)

    # meta file name
    META_FILE = '__meta__.md'

    # meta directive: rename directory name when generating site.
    META_DIRECTORY_NAME = 'directory-name'

    def __init__(self):
        self.meta_mapping = {}

    def _get_replaced_url(self, href):
        if href.startswith('http') or href.startswith('//'):
            return href
        # パスの長い順にマッチングして最初に見つかったパターンで置換, 終了
        meta_dir_keys = sorted(self.meta_mapping.keys())
        meta_dir_keys.reverse()
        for meta_dir in meta_dir_keys:
            encoded_url = urllib.parse.quote(meta_dir)
            encoded_url = encoded_url.lstrip('/')
            encoded_url = encoded_url.rstrip('/')
            if href.startswith(encoded_url):
                new_url = urllib.parse.quote(self._get_directory_name(meta_dir))
                return href.replace(encoded_url + '/', new_url + '/')
        return href

    def _get_directory_name(self, original_dir_name):
        return self.meta_mapping[original_dir_name].get(DirectoryMetaInfoPlugin.META_DIRECTORY_NAME)

    def on_files(self, files, config):
        result = []
        for f in files:
            filename = path.basename(f.src_path)
            if DirectoryMetaInfoPlugin.META_FILE == filename:
                dirname = path.dirname(f.src_path)
                if dirname == "":
                    print("warning: meta file on root directory is not supported, ignored.")
                else:
                    with open(path.join(config.get('docs_dir'), f.src_path), 'r', encoding='utf8') as meta_file:
                        content = ''.join(meta_file.readlines())
                    meta = Meta.get_data(content)
                    self.meta_mapping[path.dirname(f.src_path)] = meta[1]
                continue
            result.append(f)
        return Files(result)

    def on_pre_page(self, page, config, files):
        page.file.url = self._get_replaced_url(page.url)
        return page

    def on_page_content(self, html, page, config, files):
        replaced = ''
        position = 0
        all_found = [found for found in DirectoryMetaInfoPlugin.LINK_PATTERN.finditer(html)]
        for i, found in enumerate(all_found):
            is_last = i == len(all_found) - 1
            found_position = found.regs[1]
            replaced += html[position:found_position[0]]
            replaced += self._get_replaced_url(html[found_position[0]:found_position[1]])
            if is_last:
                replaced += html[found_position[1]:]
            else:
                next_start_pos = all_found[i + 1].regs[1][0]
                replaced += html[found_position[1]:next_start_pos]
                position = next_start_pos
        return replaced

    def on_post_build(self, config):
        def _get_path(pathname):
            site_dir = config['site_dir']
            return path.join(site_dir, pathname.lstrip('/'))

        meta_dir_keys = sorted(self.meta_mapping.keys())
        meta_dir_keys.reverse()
        for meta_dir in meta_dir_keys:
            dir_from = _get_path(meta_dir)
            dir_dest = self._get_directory_name(meta_dir)
            if dir_dest != None:
                rename(dir_from, path.join(path.dirname(dir_from), dir_dest))
